package consul

import (
	"context"
	"encoding/base64"
	"fmt"
	"time"

	"github.com/hashicorp/vault/sdk/framework"
	"github.com/hashicorp/vault/sdk/logical"
)

func pathListRoles(b *backend) *framework.Path {
	return &framework.Path{
		Pattern: "roles/?$",

		Callbacks: map[logical.Operation]framework.OperationFunc{
			logical.ListOperation: b.pathRoleList,
		},
	}
}

func pathRoles(b *backend) *framework.Path {
	return &framework.Path{
		Pattern: "roles/" + framework.GenericNameRegex("name"),
		Fields: map[string]*framework.FieldSchema{
			"name": {
				Type:        framework.TypeString,
				Description: "Name of the role.",
			},

			"policy": {
				Type: framework.TypeString,
				Description: `Policy document, base64 encoded. Required
for 'client' tokens. Required for Consul pre-1.4.`,
			},

			"policies": {
				Type: framework.TypeCommaStringSlice,
				Description: `List of policies to attach to the token. Either "policies"
or "consul_roles" are required for Consul 1.5 and above, or just "policies" if
using Consul 1.4.`,
			},

			"consul_roles": {
				Type: framework.TypeCommaStringSlice,
				Description: `List of Consul roles to attach to the token. Either "policies"
or "consul_roles" are required for Consul 1.5 and above.`,
			},

			"local": {
				Type: framework.TypeBool,
				Description: `Indicates that the token should not be replicated globally 
and instead be local to the current datacenter. Available in Consul 1.4 and above.`,
			},

			"token_type": {
				Type:    framework.TypeString,
				Default: "client",
				Description: `Which type of token to create: 'client' or 'management'. If
a 'management' token, the "policy", "policies", and "consul_roles" parameters are not
required. Defaults to 'client'.`,
			},

			"ttl": {
				Type:        framework.TypeDurationSecond,
				Description: "TTL for the Consul token created from the role.",
			},

			"max_ttl": {
				Type:        framework.TypeDurationSecond,
				Description: "Max TTL for the Consul token created from the role.",
			},

			"lease": {
				Type:        framework.TypeDurationSecond,
				Description: `Use "ttl" instead.`,
				Deprecated:  true,
			},

			"consul_namespace": {
				Type: framework.TypeString,
				Description: `Indicates which namespace that the token will be
created within. Defaults to 'default'. Available in Consul 1.7 and above.`,
			},

			"partition": {
				Type: framework.TypeString,
				Description: `Indicates which admin partition that the token
will be created within. Defaults to 'default'. Available in Consul 1.11 and above.`,
			},
		},

		Callbacks: map[logical.Operation]framework.OperationFunc{
			logical.ReadOperation:   b.pathRolesRead,
			logical.UpdateOperation: b.pathRolesWrite,
			logical.DeleteOperation: b.pathRolesDelete,
		},
	}
}

func (b *backend) pathRoleList(ctx context.Context, req *logical.Request, d *framework.FieldData) (*logical.Response, error) {
	entries, err := req.Storage.List(ctx, "policy/")
	if err != nil {
		return nil, err
	}

	return logical.ListResponse(entries), nil
}

func (b *backend) pathRolesRead(ctx context.Context, req *logical.Request, d *framework.FieldData) (*logical.Response, error) {
	name := d.Get("name").(string)

	entry, err := req.Storage.Get(ctx, "policy/"+name)
	if err != nil {
		return nil, err
	}
	if entry == nil {
		return nil, nil
	}

	var roleConfigData roleConfig
	if err := entry.DecodeJSON(&roleConfigData); err != nil {
		return nil, err
	}

	if roleConfigData.TokenType == "" {
		roleConfigData.TokenType = "client"
	}

	// Generate the response
	resp := &logical.Response{
		Data: map[string]interface{}{
			"lease":            int64(roleConfigData.TTL.Seconds()),
			"ttl":              int64(roleConfigData.TTL.Seconds()),
			"max_ttl":          int64(roleConfigData.MaxTTL.Seconds()),
			"token_type":       roleConfigData.TokenType,
			"local":            roleConfigData.Local,
			"consul_namespace": roleConfigData.ConsulNamespace,
			"partition":        roleConfigData.Partition,
		},
	}
	if roleConfigData.Policy != "" {
		resp.Data["policy"] = base64.StdEncoding.EncodeToString([]byte(roleConfigData.Policy))
	}
	if len(roleConfigData.Policies) > 0 {
		resp.Data["policies"] = roleConfigData.Policies
	}
	if len(roleConfigData.ConsulRoles) > 0 {
		resp.Data["consul_roles"] = roleConfigData.ConsulRoles
	}
	return resp, nil
}

func (b *backend) pathRolesWrite(ctx context.Context, req *logical.Request, d *framework.FieldData) (*logical.Response, error) {
	tokenType := d.Get("token_type").(string)
	policy := d.Get("policy").(string)
	policies := d.Get("policies").([]string)
	roles := d.Get("consul_roles").([]string)

	switch tokenType {
	case "client":
		if policy == "" && len(policies) == 0 && len(roles) == 0 {
			return logical.ErrorResponse(
				"Use either a policy document, a list of policies, or a list of roles, depending on your Consul version"), nil
		}
	case "management":
	default:
		return logical.ErrorResponse(
			"token_type must be \"client\" or \"management\""), nil
	}

	policyRaw, err := base64.StdEncoding.DecodeString(policy)
	if err != nil {
		return logical.ErrorResponse(fmt.Sprintf(
			"Error decoding policy base64: %s", err)), nil
	}

	var ttl time.Duration
	ttlRaw, ok := d.GetOk("ttl")
	if ok {
		ttl = time.Second * time.Duration(ttlRaw.(int))
	} else {
		leaseParamRaw, ok := d.GetOk("lease")
		if ok {
			ttl = time.Second * time.Duration(leaseParamRaw.(int))
		}
	}

	var maxTTL time.Duration
	maxTTLRaw, ok := d.GetOk("max_ttl")
	if ok {
		maxTTL = time.Second * time.Duration(maxTTLRaw.(int))
	}

	name := d.Get("name").(string)
	local := d.Get("local").(bool)
	namespace := d.Get("consul_namespace").(string)
	partition := d.Get("partition").(string)
	entry, err := logical.StorageEntryJSON("policy/"+name, roleConfig{
		Policy:          string(policyRaw),
		Policies:        policies,
		ConsulRoles:     roles,
		TokenType:       tokenType,
		TTL:             ttl,
		MaxTTL:          maxTTL,
		Local:           local,
		ConsulNamespace: namespace,
		Partition:       partition,
	})
	if err != nil {
		return nil, err
	}

	if err := req.Storage.Put(ctx, entry); err != nil {
		return nil, err
	}

	return nil, nil
}

func (b *backend) pathRolesDelete(ctx context.Context, req *logical.Request, d *framework.FieldData) (*logical.Response, error) {
	name := d.Get("name").(string)
	if err := req.Storage.Delete(ctx, "policy/"+name); err != nil {
		return nil, err
	}
	return nil, nil
}

type roleConfig struct {
	Policy          string        `json:"policy"`
	Policies        []string      `json:"policies"`
	ConsulRoles     []string      `json:"consul_roles"`
	TTL             time.Duration `json:"lease"`
	MaxTTL          time.Duration `json:"max_ttl"`
	TokenType       string        `json:"token_type"`
	Local           bool          `json:"local"`
	ConsulNamespace string        `json:"consul_namespace"`
	Partition       string        `json:"partition"`
}
